from __future__ import absolute_import

import csv
import io


def get_csv_rows_as_dicts(csv_filepath, csv_field_meta=None):
    """
    Generic method that generates a dictionary for each line in a CSV
    file using the CSV headers as keys.
    """
    with io.open(csv_filepath, encoding='utf-8') as f:
        for row_dict in csv.DictReader(f):
            if not csv_field_meta:
                yield row_dict
            else:
                yield {
                    k: csv_field_meta[k]['validator'](row_dict[csv_field_meta[k]['csv_header']]) for k in csv_field_meta
                }



# -*- coding: utf-8 -*-

import io
import json
import os
import shutil

from argparse import RawDescriptionHelpFormatter

from oasislmf.model_execution.bash import genbash
from ..exposures.manager import OasisExposuresManager
from ..model_execution.runner import run
from ..model_execution.bin import create_binary_files, prepare_model_run_directory, prepare_model_run_inputs
from ..utils.exceptions import OasisException
from ..utils.values import get_utctimestamp
from ..keys.lookup import OasisKeysLookupFactory
from .cleaners import PathCleaner, as_path
from .base import OasisBaseCommand, InputValues


class GenerateKeysCmd(OasisBaseCommand):
    """
    Generate and write Oasis keys (area peril ID, vulnerability ID) for a model.

    The command line arguments can be supplied in the configuration file
    (``oasislmf.json`` by default or specified with the ``--config`` flag).
    Run ``oasislmf config --help`` for more information.

    Keys records returned by an Oasis keys lookup service (see the PiWind
    lookup service for reference) will be Python dicts with the following
    structure
    ::

        {
            "id": <loc. ID>,
            "peril_id": <Oasis peril type ID - oasis_utils/oasis_utils.py>,
            "coverage": <Oasis coverage type ID - see oasis_utils/oasis_utils.py>,
            "area_peril_id": <area peril ID>,
            "vulnerability_id": <vulnerability ID>,
            "message": <lookup status message>,
            "status": <lookup status code - see oasis_utils/oasis_utils.py>
        }

    The script can generate keys records in this format, and write them to file.

    For model loss calculations however ``ktools`` requires a keys CSV file with
    the following format
    ::

        LocID,PerilID,CoverageID,AreaPerilID,VulnerabilityID
        ..
        ..

    where the headers correspond to the relevant Oasis keys record fields.
    The script can also generate and write Oasis keys files.
    """
    formatter_class = RawDescriptionHelpFormatter

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(GenerateKeysCmd, self).add_args(parser)

        parser.add_argument(
            'output-file-path', type=PathCleaner('Output file', preexists=False),
            help='Keys records output file path',
        )
        parser.add_argument(
            '-k', '--keys-data-path', default=None,
            help='Keys data directory path',
        )
        parser.add_argument(
            '-v', '--model-version-file-path', default=None,
            help='Model version file path',
        )
        parser.add_argument(
            '-l', '--lookup-package-path', default=None,
            help='Keys data directory path',
        )
        parser.add_argument(
            '-t', '--output-format', choices=['oasis_keys', 'list_keys'],
            help='Keys records file output format',
        )
        parser.add_argument(
            '-e', '--model-exposures-file-path', default=None, help='Keys records file output format',
        )
        parser.add_argument(
            '-s', '--successes-only', action='store_true', help='Only record successful entries',
        )
        parser.set_defaults(successes_only=False)

    def action(self, args):
        """
        Generate and write Oasis keys (area peril ID, vulnerability ID) for a model.

        :param args: The arguments from the command line
        :type args: Namespace
        """
        inputs = InputValues(args)
        model_exposures_file_path = as_path(inputs.get('model_exposures_file_path', required=True, is_path=True), 'Model exposures')
        keys_data_path = as_path(inputs.get('keys_data_path', required=True, is_path=True), 'Keys data')
        version_file_path = as_path(inputs.get('model_version_file_path', required=True, is_path=True), 'Version file')
        lookup_package_path = as_path(inputs.get('lookup_package_path', required=True, is_path=True), 'Lookup package')

        self.logger.info('Getting model info and creating lookup service instance')
        model_info, model_klc = OasisKeysLookupFactory.create(
            model_keys_data_path=keys_data_path,
            model_version_file_path=version_file_path,
            lookup_package_path=lookup_package_path,
        )
        self.logger.info('\t{}, {}'.format(model_info, model_klc))

        self.logger.info('Saving keys records to file')
        f, n = OasisKeysLookupFactory.save_keys(
            lookup=model_klc,
            model_exposures_file_path=model_exposures_file_path,
            output_file_path=args.output_file_path,
            success_only=args.successes_only
        )
        self.logger.info('{} keys records saved to file {}'.format(n, f))


class GenerateOasisFilesCmd(OasisBaseCommand):
    """
    Generate Oasis files (items, coverages, GUL summary) for a model

    The command line arguments can be supplied in the configuration file
    (``oasislmf.json`` by default or specified with the ``--config`` flag).
    """
    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(GenerateOasisFilesCmd, self).add_args(parser)

        parser.add_argument('oasis_files_path', default=None, help='Path to Oasis files', nargs='?')
        parser.add_argument('-k', '--keys-data-path', default=None, help='Path to Oasis files')
        parser.add_argument('-v', '--model-version-file-path', default=None, help='Model version file path')
        parser.add_argument('-l', '--lookup-package-path', default=None, help='Keys data directory path')
        parser.add_argument(
            '-p', '--canonical-exposures-profile-json-path', default=None,
            help='Path of the supplier canonical exposures profile JSON file'
        )
        parser.add_argument('-e', '--source-exposures-file-path', default=None, help='Source exposures file path')
        parser.add_argument(
            '-a', '--source-exposures-validation-file-path', default=None,
            help='Source exposures validation file (XSD) path'
        )
        parser.add_argument(
            '-b', '--source-to-canonical-exposures-transformation-file-path', default=None,
            help='Source -> canonical exposures transformation file (XSLT) path'
        )
        parser.add_argument(
            '-c', '--canonical-exposures-validation-file-path', default=None,
            help='Canonical exposures validation file (XSD) path'
        )
        parser.add_argument(
            '-d', '--canonical-to-model-exposures-transformation-file-path', default=None,
            help='Canonical exposures validation file (XSD) path'
        )

    def action(self, args):
        """
        Generate Oasis files (items, coverages, GUL summary) for a model

        :param args: The arguments from the command line
        :type args: Namespace
        """
        inputs = InputValues(args)

        utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')
        default_oasis_files_path = os.path.join(os.getcwd(), 'runs', 'OasisFiles-{}'.format(utcnow))
        oasis_files_path = as_path(inputs.get('oasis_files_path', is_path=True, default=default_oasis_files_path), 'Oasis file', preexists=False)
        keys_data_path = as_path(inputs.get('keys_data_path', required=True, is_path=True), 'Keys data')
        model_version_file_path = as_path(inputs.get('model_version_file_path', required=True, is_path=True), 'Model version file')
        lookup_package_file_path = as_path(inputs.get('lookup_package_path', required=True, is_path=True), 'Lookup package file')
        canonical_exposures_profile_json_path = as_path(
            inputs.get('canonical_exposures_profile_json_path', required=True, is_path=True),
            'Canonical exposures profile json'
        )
        source_exposures_file_path = as_path(inputs.get('source_exposures_file_path', required=True, is_path=True), 'Source exposures')
        source_exposures_validation_file_path = as_path(
            inputs.get('source_exposures_validation_file_path', required=True, is_path=True),
            'Source exposures validation file'
        )
        source_to_canonical_exposures_transformation_file_path = as_path(
            inputs.get('source_to_canonical_exposures_transformation_file_path', required=True, is_path=True),
            'Source to canonical exposures transformation'
        )
        canonical_exposures_validation_file_path = as_path(
            inputs.get('canonical_exposures_validation_file_path', required=True, is_path=True),
            'Canonical exposures validation file'
        )
        canonical_to_model_exposures_transformation_file_path = as_path(
            inputs.get('canonical_to_model_exposures_transformation_file_path', required=True, is_path=True),
            'Canonical to model exposures transformation file'
        )

        self.logger.info('Getting model info and creating lookup service instance')
        model_info, model_klc = OasisKeysLookupFactory.create(
            model_keys_data_path=keys_data_path,
            model_version_file_path=model_version_file_path,
            lookup_package_path=lookup_package_file_path,
        )
        self.logger.info('\t{}, {}'.format(model_info, model_klc))

        self.logger.info('Creating Oasis model object')
        model = OasisExposuresManager().create(
            model_supplier_id=model_info['supplier_id'],
            model_id=model_info['model_id'],
            model_version_id=model_info['model_version_id'],
            resources={
                'lookup': model_klc,
                'oasis_files_path': oasis_files_path,
                'canonical_exposures_profile_json_path': canonical_exposures_profile_json_path,
                'source_exposures_validation_file_path': source_exposures_validation_file_path,
                'source_to_canonical_exposures_transformation_file_path': source_to_canonical_exposures_transformation_file_path,
                'canonical_exposures_validation_file_path': canonical_exposures_validation_file_path,
                'canonical_to_model_exposures_transformation_file_path': canonical_to_model_exposures_transformation_file_path,
            }
        )
        self.logger.info('\t{}'.format(model))

        self.logger.info('Setting up Oasis files directory for model {}'.format(model.key))
        if not os.path.exists(oasis_files_path):
            os.mkdir(oasis_files_path)

        self.logger.info('Generating Oasis files for model')
        oasis_files = OasisExposuresManager().start_files_pipeline(
            oasis_model=model,
            oasis_files_path=oasis_files_path,
            source_exposures_path=source_exposures_file_path,
            logger=self.logger,
        )

        self.logger.info('Generated Oasis files for model: {}'.format(oasis_files))


class GenerateLossesCmd(OasisBaseCommand):
    """
    Generate losses using the installed ktools framework.

    Given Oasis files, model analysis settings JSON file, model data, and
    some other parameters. can generate losses using the installed ktools framework.

    The command line arguments can be supplied in the configuration file
    (``oasislmf.json`` by default or specified with the ``--config`` flag).
    Run ``oasislmf config --help`` for more information.

    The script creates a time-stamped folder in the model run directory and
    sets that as the new model run directory, copies the analysis settings
    JSON file into the run directory and creates the following folder
    structure
    ::

        ├── analysis_settings.json
        ├── fifo/
        ├── input/
        ├── output/
        ├── static/
        └── work/

    Depending on the OS type the model data is symlinked (Linux, Darwin) or
    copied (Cygwin, Windows) into the ``static`` subfolder. The input files
    are kept in the ``input`` subfolder and the losses are generated as CSV
    files in the ``output`` subfolder.

    By default executing the generated ktools losses script will automatically
    execute, this can be overridden by providing the ``--no-execute`` flag.
    """
    formatter_class = RawDescriptionHelpFormatter

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(GenerateLossesCmd, self).add_args(parser)

        parser.add_argument('oasis_files_path', default=None, help='Path to Oasis files', nargs='?')
        parser.add_argument('-j', '--analysis-settings-json-file-path', default=None, help='Relative or absolute path of the model analysis settings JSON file')
        parser.add_argument('-m', '--model-data-path', default=None, help='Model data source path')
        parser.add_argument('-r', '--model-run-dir-path', default=None, help='Model run directory path')
        parser.add_argument('-s', '--ktools-script-name', default=None, help='Relative or absolute path of the output file')
        parser.add_argument('-n', '--ktools-num-processes', default=-1, help='Number of ktools calculation processes to use')
        parser.add_argument('-x', '--no-execute', action='store_true', help='Whether to execute generated ktools script')

    def action(self, args):
        """
        Generate losses using the installed ktools framework.

        :param args: The arguments from the command line
        :type args: Namespace
        """
        inputs = InputValues(args)

        oasis_files_path = as_path(inputs.get('oasis_files_path', required=True, is_path=True), 'Oasis files', preexists=True)

        model_run_dir_path = as_path(inputs.get('model_run_dir_path', required=False, is_path=True), 'Model run directory', preexists=False)

        analysis_settings_json_file_path = as_path(
            inputs.get('analysis_settings_json_file_path', required=True, is_path=True),
            'Analysis settings file'
        )
        model_data_path = as_path(inputs.get('model_data_path', required=True, is_path=True), 'Model data')
        
        ktools_script_name = inputs.get('ktools_script_name', default='run_ktools')
        no_execute = inputs.get('no_execute', default=False)
        
        if not model_run_dir_path:
            utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')
            model_run_dir_path = os.path.join(os.getcwd(), 'runs', 'ProgOasis-{}'.format(utcnow))
            self.logger.info('No model run dir. provided - creating a timestamped run dir. in working directory as {}'.format(model_run_dir_path))
            os.mkdir(model_run_dir_path)
        else:
            if not os.path.exists(model_run_dir_path):
                os.mkdir(model_run_dir_path)

        self.logger.info(
            'Preparing model run directory {} - copying Oasis files, analysis settings JSON file and linking model data'.format(model_run_dir_path)
        )
        prepare_model_run_directory(
            model_run_dir_path,
            oasis_files_path,
            analysis_settings_json_file_path,
            model_data_path
        )

        self.logger.info('Converting Oasis files to ktools binary files')
        oasis_files_path = os.path.join(model_run_dir_path, 'input', 'csv')
        binary_files_path = os.path.join(model_run_dir_path, 'input')
        create_binary_files(oasis_files_path, binary_files_path)

        analysis_settings_json_file_path = os.path.join(model_run_dir_path, 'analysis_settings.json')
        try:
            self.logger.info('Reading analysis settings JSON file')
            with io.open(analysis_settings_json_file_path, 'r', encoding='utf-8') as f:
                analysis_settings = json.load(f)
                if 'analysis_settings' in analysis_settings:
                    analysis_settings = analysis_settings['analysis_settings']
        except (IOError, TypeError, ValueError):
            raise OasisException('Invalid analysis settings JSON file or file path: {}.'.format(analysis_settings_json_file_path))

        self.logger.info('Loaded analysis settings JSON: {}'.format(analysis_settings))

        self.logger.info('Preparing model run inputs')
        prepare_model_run_inputs(analysis_settings, model_run_dir_path)

        script_path = os.path.join(model_run_dir_path, '{}.sh'.format(ktools_script_name))
        if no_execute:
            self.logger.info('Generating ktools losses script')
            genbash(
                args.ktools_num_processes,
                analysis_settings,
                filename=script_path,
            )
            self.logger.info('Making ktools losses script executable')
            subprocess.check_call("chmod +x {}".format(script_path), stderr=subprocess.STDOUT, shell=True)
        else:
            os.chdir(model_run_dir_path)
            run(analysis_settings, args.ktools_num_processes, filename=script_path)

        self.logger.info('Loss outputs generated in {}'.format(os.path.join(model_run_dir_path, 'output')))


class RunCmd(OasisBaseCommand):
    """
    Run models end to end.

    The command line arguments can be supplied in the configuration file
    (``oasislmf.json`` by default or specified with the ``--config`` flag).
    """
    def add_args(self, parser):
        """
        Run models end to end.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(RunCmd, self).add_args(parser)

        parser.add_argument('-k', '--keys-data-path', default=None, help='Path to Oasis files')
        parser.add_argument('-v', '--model-version-file-path', default=None, help='Model version file path')
        parser.add_argument('-l', '--lookup-package-file-path', default=None, help='Keys data directory path')
        parser.add_argument(
            '-p', '--canonical-exposures-profile-json-path', default=None,
            help='Path of the supplier canonical exposures profile JSON file'
        )
        parser.add_argument('-e', '--source-exposures-file-path', default=None, help='Source exposures file path')
        parser.add_argument(
            '-a', '--source-exposures-validation-file-path', default=None,
            help='Source exposures validation file (XSD) path'
        )
        parser.add_argument(
            '-b', '--source-to-canonical-exposures-transformation-file-path', default=None,
            help='Source -> canonical exposures transformation file (XSLT) path'
        )
        parser.add_argument(
            '-c', '--canonical-exposures-validation-file-path', default=None,
            help='Canonical exposures validation file (XSD) path'
        )
        parser.add_argument(
            '-d', '--canonical-to-model-exposures-transformation-file-path', default=None,
            help='Canonical exposures validation file (XSD) path'
        )
        parser.add_argument(
            '-j', '--analysis-settings-json-file-path', default=None,
            help='Model analysis settings JSON file path'
        )
        parser.add_argument('-m', '--model-data-path', default=None, help='Model data source path')
        parser.add_argument('-r', '--model-run-dir-path', default=None, help='Model run directory path')
        parser.add_argument(
            '-s', '--ktools-script-name', default=None,
            help='Name of the ktools output script (should not contain any filetype extension)'
        )
        parser.add_argument('-n', '--ktools-num-processes', default=2, help='Number of ktools calculation processes to use')

    def action(self, args):
        """
        Generate Oasis files (items, coverages, GUL summary) for a model

        :param args: The arguments from the command line
        :type args: Namespace
        """
        inputs = InputValues(args)
        model_run_dir_path = as_path(inputs.get('model_run_dir_path', required=False), 'Model run path', preexists=False)

        if not model_run_dir_path:
            utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')
            model_run_dir_path = os.path.join(os.getcwd(), 'runs', 'ProgOasis-{}'.format(utcnow))
            self.logger.info('No model run dir. provided - creating a timestamped run dir. in working directory as {}'.format(model_run_dir_path))
            os.mkdir(model_run_dir_path)
        else:
            if not os.path.exists(model_run_dir_path):
                os.mkdir(model_run_dir_path)

        args.model_run_dir_path = model_run_dir_path

        args.oasis_files_path = os.path.join(model_run_dir_path, 'tmp')
        self.logger.info('Creating temporary folder {} for Oasis files'.format(args.oasis_files_path))
        os.mkdir(args.oasis_files_path)

        gen_oasis_files_cmd = GenerateOasisFilesCmd()
        gen_oasis_files_cmd._logger = self.logger
        gen_oasis_files_cmd.action(args)

        gen_losses_cmd = GenerateLossesCmd()
        gen_losses_cmd._logger = self.logger
        gen_losses_cmd.action(args)


class ModelsCmd(OasisBaseCommand):
    sub_commands = {
        'generate-keys': GenerateKeysCmd,
        'generate-oasis-files': GenerateOasisFilesCmd,
        'generate-losses': GenerateLossesCmd,
        'run': RunCmd,
    }



# -*- coding: utf-8 -*-
import json
import logging

__all__ = [
    'Translator'
]

import csv
import io
import os
import pandas as pd

from itertools import islice

from lxml import etree


class Translator(object):
    def __init__(self, input_path, output_path, xsd_path, xslt_path, append_row_nums=False, chunk_size=5000, logger=None):
        self.logger = logger or logging.getLogger()
        self.threshold = 100000000
        # Max file size for file_ReadAll() method, in bytes [100MB]

        self.xsd = etree.parse(xsd_path)   # validation file
        self.xslt = etree.parse(xslt_path)  # transform file
        self.fpath_input = input_path        # file_in.csv
        self.fpath_output = output_path       # file_out.csv
        self.ext = os.path.splitext(self.fpath_output)[1]

        # Row Control Vars
        self.row_nums = append_row_nums  # Is 's' or 'S' as well?
        self.row_limit = chunk_size    # MAX number of CSV input rows
        self.row_start = 1             # Start of input file segment to process
        self.row_end = self.row_limit  # end of input file segment to process
        self.row_header_in = None      # CSV col header
        self.row_header_out = None     # CSV col header Post Transform

        # Input Validation
        if (self.ext == ''):
            raise TypeError("Missing file extention for output file.")


# --- Main loop --------------------------------------------------------------#

    def __call__(self):

        # read input CSV header
        with io.open(self.fpath_input, 'r') as fd_input:
            self.row_header_in = self.read_file_slice(fd_input, 0, 0)[0]

            for data, first_row_number, last_row_number in self.next_file_slice(fd_input):
                self.logger.debug('--- lines[%d .. %d] -------------------------------', first_row_number, last_row_number)

                # Convert CSV -> XML
                xml_input_slice = self.csv_to_xml(
                    self.row_header_in,
                    data
                )
                self.print_xml(xml_input_slice)

                # Transform
                xml_output = self.xml_transform(xml_input_slice, self.xslt)
                self.print_xml(xml_output)

                # Validate Output
                self.logger.info(self.xml_validate(xml_output, self.xsd))

                # Convert transform XML back to CSV
                self.xml_to_csv(
                    xml_output,          # XML etree
                    first_row_number,  # First Row in this slice
                    last_row_number    # Last Row in this slice
                )

# --- Transform Functions ----------------------------------------------------#
# https://pymotw.com/2/xml/etree/ElementTree/create.html
# http://lxml.de/api/lxml.etree._Element-class.html

    # --- CSV Funcs --- #
    def csv_to_xml(self, csv_header, csv_data):
        root = etree.Element('root')
        # fetch each row
        for row in csv_data:
            # Create new 'empty' record
            rec = etree.SubElement(root, 'rec')
            # Iter over columns and set attributs
            for i in range(0, len(row)):
                rec.set(self.row_header_in[i], row[i])
        return root

    def csv_insert_row_nums(self, csv_data, r_start, r_end):
        row_index = range(r_start, r_end)
        row_total = r_end - r_start
        if(row_total != len(csv_data)):
            raise TypeError('Size mismatch between row numbering and dataset')
        return [[row_index[i]] + csv_data[i] for i in range(0, row_total)]

    # --- XML Funcs --- #
    def xml_to_csv(self, xml_elementTree, row_first, row_last):
        root = xml_elementTree.getroot()

        # Check if this is the first file chunk processed
        # and Extract the New CSV header
        if not (self.row_header_out):
            self.row_header_out = root[0].keys()

            if self.row_nums:
                self.row_header_out.insert(0, 'ROW_ID')
            # Append first line to output csv file
            self.write_file_header(self.row_header_out)

        # Convert each chunk into Python Dict then pass to:
        #     class csv.DictWriter(csvfile ... )
        #     see: https://docs.python.org/2/library/csv.html

        line_counter = row_first

        for rec in root:
            # print([rec.get(key) for key in self.row_header_out])

            # Convert Row record to python dict() Object
            rec_d = rec.attrib

            # append ROW_ID
            if self.row_nums:
                rec_d['ROW_ID'] = str(line_counter)
                line_counter += 1

            # Append to output file
            self.print_dict(rec_d)
            self.append_file_row(self.row_header_out, rec_d)

        # guard for correct Row numbering
        if ((self.row_nums) and (line_counter != row_last)):
            self.logger.error('Line_count: %d, Row_last: %d', line_counter, row_last)

            raise KeyError('Row_ID missmatch')

    # http://lxml.de/2.0/validation.html
    # If valid   -> Return True
    #    invalid -> error_log
    def xml_validate(self, xml_etree, xsd_etree):
        xmlSchema = etree.XMLSchema(xsd_etree)
        self.print_xml(xml_etree)
        self.print_xml(xsd_etree)
        # Calling 'assertValid' will raise execptions,
        # --> should be handled above this level
        # if (xmlSchema.assertValid(xml_etree)):
        if (xmlSchema.validate(xml_etree)):
            return True
        else:
            self.logger.error('Input failed to Validate')
            log = xmlSchema.error_log
            self.logger.error(log.last_error)
            return False

    # http://lxml.de/xpathxslt.html#xslt
    def xml_transform(self, xml_doc, xslt):
        lxml_transform = etree.XSLT(self.xslt)
        return lxml_transform(xml_doc)


# --- File I/O Functions -----------------------------------------------------#

    # Generator Function which processes and returns batches of the input CSV
    def next_file_slice(self, file_object):
        while True:
            csv_chunk = self.read_file_slice(
                file_object,
                self.row_start,
                self.row_end
            )

            # Exit check for EOF
            rowsInChunk = len(csv_chunk)

            if (rowsInChunk == 0):
                break
            else:
                slice_start = self.row_start
                slice_end = self.row_start + rowsInChunk

                self.row_start += self.row_limit
                self.row_end += self.row_limit
                yield csv_chunk, slice_start, slice_end

    # Return Line numbers of a file between [l_start .. l_end]
    def read_file_slice(self, file_obj, l_start, l_end):
        # return pointer to start of file
        file_obj.seek(0)
        input_reader = csv.reader(file_obj, delimiter=',')
        # create iterator for the file slice
        file_slice = islice(input_reader, l_start, l_end + 1)
        return [line for line in file_slice]

    # Function to append output as its processed in batches
    #
    # file_object: OutputFile
    # row_names: list of row names  -->   ['ROW_ID', 'ACCNTNUM', .. ]
    # row_data: dict mapping for XML Atribuites:
    # {'ROW_ID': '1', 'ACCNTNUM': '0.02',   ....  }
    # https://docs.python.org/2/library/csv.html
    def append_file_row(self, row_names, row_data):
        pd.DataFrame(columns=row_names, data={k: [v] for k, v in row_data.items()}).to_csv(
            self.fpath_output,
            mode='a',
            encoding='utf-8',
            header=False,
            index=False,
        )

    # WARNING: This will overwrite the file path
    def write_file_header(self, row_names):
        pd.DataFrame(columns=row_names).to_csv(
            self.fpath_output,
            encoding='utf-8',
            header=True,
            index=False,
        )

# --- Verbose Print Funcs ----------------------------------------------------#

    def print_dict(self, d):
        self.logger.debug(json.dumps(dict(d), indent=4))

    def print_xml(self, etree_obj):
        self.logger.debug('___________________________________________')
        self.logger.debug(etree.tostring(etree_obj, pretty_print=True))



from argparse import RawDescriptionHelpFormatter

from .base import OasisBaseCommand


class ConfigCmd(OasisBaseCommand):
    """
    The ``oasislmf`` tool can be configured using using
    a json configuration file. By default this is stored
    in ``oasislmf.json`` but the path can be changed by
    specifying the ``--config`` flag.

    This configuration file should contain any or all of
    the following properties in no particular order
    ::

        keys_data_path
        model_data_path
        model_version_file_path
        model_exposures_file_path
        model_run_dir_path
        output_file_path
        output_format
        lookup_package_path
        analysis_settings_json_file_path
        ktools_script_name
        ktools_num_processes
        no_execute
        canonical_exposures_profile_json_path
        canonical_exposures_validation_file_path
        canonical_to_model_exposures_transformation_file_path
        source_exposures_file_path
        source_exposures_validation_file_path
        source_to_canonical_exposures_transformation_file_path

    The path-related keys should be string paths, given relative
    to the location of JSON file.

    As an example, this is the master script configuration file for PiWind

    ::

        {
            "keys_data_path": "OasisPiWind/keys_data/PiWind",
            "model_version_file_path": "OasisPiWind/keys_data/PiWind/ModelVersion.csv",
            "lookup_package_path": "OasisPiWind/src/keys_server",
            "canonical_exposures_profile_json_path": "OasisPiWind/oasislmf-piwind-canonical-profile.json",
            "source_exposures_file_path": "OasisPiWind/tests/data/SourceLocPiWind.csv",
            "source_exposures_validation_file_path": "OasisPiWind/flamingo/PiWind/Files/ValidationFiles/Generic_Windstorm_SourceLoc.xsd",
            "source_to_canonical_exposures_transformation_file_path": "OasisPiWind/flamingo/PiWind/Files/TransformationFiles/MappingMapToGeneric_Windstorm_CanLoc_A.xslt",
            "canonical_exposures_validation_file_path": "OasisPiWind/flamingo/PiWind/Files/ValidationFiles/Generic_Windstorm_CanLoc_B.xsd",
            "canonical_to_model_exposures_transformation_file_path": "OasisPiWind/flamingo/PiWind/Files/TransformationFiles/MappingMapTopiwind_modelloc.xslt",
            "xtrans_path": "omdk/xtrans/xtrans.exe",
            "oasis_files_path": "omdk/runs",
            "analysis_settings_json_file_path": "OasisPiWind/analysis_settings.json",
            "model_data_path": "OasisPiWind/model_data/PiWind",
            "model_run_dir_path": "omdk/runs"
        }

    It can also be obtained from
    `https://github.com/OasisLMF/OasisPiWind/blob/master/mdk-oasislmf-piwind.json <https://github.com/OasisLMF/OasisPiWind/blob/master/mdk-oasislmf-piwind.json>`_.
    """
    formatter_class = RawDescriptionHelpFormatter



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

__all__ = [
    'OasisKeysLookupFactory',
    'OasisBaseKeysLookup',
]

import csv
import importlib
import io
import json
import os
import re
import sys

from collections import OrderedDict

import pandas as pd

from six import StringIO

from ..utils.exceptions import OasisException
from ..utils.log import oasis_log
from ..utils.status import KEYS_STATUS_NOMATCH, KEYS_STATUS_SUCCESS

UNKNOWN_ID = -1


class OasisBaseKeysLookup(object):  # pragma: no cover
    """
    A base class / interface that serves a template for model-specific keys
    lookup classes.
    """
    @oasis_log()
    def __init__(
        self,
        keys_data_directory=None,
        supplier=None,
        model_name=None,
        model_version=None
    ):
        """
        Class constructor
        """
        if keys_data_directory is not None:
            self.keys_data_directory = keys_data_directory
        else:
            self.keys_data_directory = os.path.join(os.sep, 'var', 'oasis', 'keys_data')

        self.supplier = supplier
        self.model_name = model_name
        self.model_version = model_version

    @oasis_log()
    def process_locations(self, loc_df):
        """
        Process location rows - passed in as a pandas dataframe.
        """
        pass

    def _get_location_record(self, raw_loc_item):
        """
        Returns a dict of standard location keys and values based on
        a raw location item, which is a row in a Pandas dataframe.
        """
        pass

    def _get_area_peril_id(self, record):
        """
        Get the area peril ID for a particular location record.
        """
        return UNKNOWN_ID, "Not implemented"

    def _get_vulnerability_id(self, record):
        """
        Get the vulnerability ID for a particular location record.
        """
        return UNKNOWN_ID, "Not implemented"

    @oasis_log()
    def _get_area_peril_ids(self, loc_data, include_context=True):
        """
        Generates area peril IDs in two modes - if include_context is
        True (default) it will generate location records/rows including
        the area peril IDs, otherwise it will generate pairs of location
        IDs and the corresponding area peril IDs.
        """
        pass

    @oasis_log()
    def _get_vulnerability_ids(self, loc_data, include_context=True):
        """
        Generates vulnerability IDs in two modes - if include_context is
        True (default) it will generate location records/rows including
        the area peril IDs, otherwise it will generate pairs of location
        IDs and the corresponding vulnerability IDs.
        """
        pass

    def _get_lookup_success(self, ap_id, vul_id):
        """
        Determine the status of the keys lookup.
        """
        if ap_id == UNKNOWN_ID or vul_id == UNKNOWN_ID:
            return KEYS_STATUS_NOMATCH
        return KEYS_STATUS_SUCCESS


class OasisKeysLookupFactory(object):
    """
    A factory class to load and run keys lookup services for different
    models/suppliers.
    """
    @classmethod
    def get_model_info(cls, model_version_file_path):
        """
        Get model information from the model version file.
        """
        with io.open(model_version_file_path, 'r', encoding='utf-8') as f:
            return next(csv.DictReader(
                f, fieldnames=['supplier_id', 'model_id', 'model_version_id']
            ))

    @classmethod
    def get_lookup_package(cls, lookup_package_path):
        """
        Returns the lookup service parent package (called `keys_server` and
        located in `src` in the model keys server Git repository or in
        `var/www/oasis` in the keys server Docker container) from the given
        path.
        """
        parent_dir = os.path.abspath(os.path.dirname(lookup_package_path))
        package_name = re.sub(r'\.py$', '', os.path.basename(lookup_package_path))

        sys.path.insert(0, parent_dir)
        lookup_package = importlib.import_module(package_name)
        sys.path.pop(0)

        return lookup_package

    @classmethod
    def get_lookup_class_instance(cls, lookup_package, keys_data_path, model_info):
        """
        Get the keys lookup class instance.
        """
        klc = getattr(lookup_package, '{}KeysLookup'.format(model_info['model_id']))

        return klc(
            keys_data_directory=keys_data_path,
            supplier=model_info['supplier_id'],
            model_name=model_info['model_id'],
            model_version=model_info['model_version_id']
        )

    @classmethod
    def get_model_exposures(cls, model_exposures=None, model_exposures_file_path=None):
        """
        Get the model exposures/location file data as a pandas dataframe given
        either the path of the model exposures file or the string contents of
        such a file.
        """
        if model_exposures_file_path:
            loc_df = pd.read_csv(os.path.abspath(model_exposures_file_path), float_precision='high')
        elif model_exposures:
            loc_df = pd.read_csv(StringIO(model_exposures), float_precision='high')
        else:
            raise OasisException('Either model_exposures_file_path or model_exposures must be specified')

        loc_df = loc_df.where(loc_df.notnull(), None)
        loc_df.columns = loc_df.columns.str.lower()

        return loc_df

    @classmethod
    def write_oasis_keys_file(cls, records, output_file_path):
        """
        Writes an Oasis keys file from an iterable of keys records.
        """
        heading_row = OrderedDict([
            ('id', 'LocID'),
            ('peril_id', 'PerilID'),
            ('coverage', 'CoverageID'),
            ('area_peril_id', 'AreaPerilID'),
            ('vulnerability_id', 'VulnerabilityID'),
        ])

        pd.DataFrame(
            columns=heading_row.keys(),
            data=[heading_row] + records,
        ).to_csv(
            output_file_path,
            index=False,
            encoding='utf-8',
            header=False,
        )

        return output_file_path, len(records)

    @classmethod
    def write_list_keys_file(cls, records, output_file_path):
        """
        Writes the keys records as a simple list to file.
        """
        with io.open(output_file_path, 'w', encoding='utf-8') as f:
            count = 0
            for count, record in enumerate(records, start=1):
                f.write('{},\n'.format(json.dumps(record, sort_keys=True, indent=4)))

            return f.name, count

    @classmethod
    def create(
        cls,
        model_keys_data_path=None,
        model_version_file_path=None,
        lookup_package_path=None,
    ):
        """
        Creates a keys lookup class instance for the given model and supplier -
        local file paths are required for the model keys data folder, the model
        version file and the Git repository for the model keys server. Returns a
        pair ``(model_info, klc)``, where ``model_info`` is a dictionary holding
        model information from the model version file and `klc` is the lookup
        service class instance for the model.
        """
        if model_keys_data_path:
            model_keys_data_path = os.path.abspath(model_keys_data_path)

        if model_version_file_path:
            model_version_file_path = os.path.abspath(model_version_file_path)

        if lookup_package_path:
            lookup_package_path = os.path.abspath(lookup_package_path)

        model_info = cls.get_model_info(model_version_file_path)
        lookup_package = cls.get_lookup_package(lookup_package_path)
        klc = cls.get_lookup_class_instance(lookup_package, model_keys_data_path, model_info)
        return model_info, klc

    @classmethod
    def get_keys(
        cls,
        lookup=None,
        model_exposures=None,
        model_exposures_file_path=None,
        success_only=True
    ):
        """
        Generates keys keys records (JSON) for the given model and supplier -
        requires an instance of the lookup service (which can be created using
        the `create` method in this factory class), and either the model
        location file path or the string contents of such a file.

        The optional keyword argument ``success_only`` indicates whether only
        records with successful lookups should be returned (default), or all
        records.
        """
        if not (model_exposures or model_exposures_file_path):
            raise OasisException('No model exposures provided')

        model_loc_df = cls.get_model_exposures(
            model_exposures_file_path=model_exposures_file_path,
            model_exposures=model_exposures,
        )

        for record in lookup.process_locations(model_loc_df):
            if success_only:
                if record['status'].lower() == 'success':
                    yield record
            else:
                yield record

    @classmethod
    def save_keys(
        cls,
        lookup=None,
        output_file_path=None,
        model_exposures=None,
        model_exposures_file_path=None,
        success_only=True
    ):
        """
        Writes the keys keys records generated by the lookup service for the
        given model and supplier to a local file - requires a lookup service
        instance (which can be created using the `create` method in this
        factory class), the path of the model location file, the path of
        output file, and the format of the output file which can be an
        Oasis keys file (``oasis_keys``) or a simple listing of the records
        to file (``list_keys``).

        The optional keyword argument ``success_only`` indicates whether only
        records with successful lookups should be returned (default), or all
        records.

        Returns a pair ``(f, n)`` where ``f`` is the output file object
        and ``n`` is the number of records written to the file.
        """
        if not (model_exposures or model_exposures_file_path):
            raise OasisException('No model exposures or model exposures file path provided')

        output_file_path = os.path.abspath(output_file_path)

        keys = cls.get_keys(
            lookup=lookup,
            model_exposures=model_exposures,
            model_exposures_file_path=model_exposures_file_path,
            success_only=success_only
        )

        return cls.write_oasis_keys_file(list(keys), output_file_path)



from __future__ import print_function, division

import glob
import os
import io
import re
import shutil
import sys
import tarfile
from contextlib import contextmanager
from distutils.log import INFO
from tempfile import mkdtemp

from setuptools import find_packages, setup
from setuptools.command.install import install

try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve

KTOOLS_VERSION = '0_0_392_0'
SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))


def get_readme():
    with io.open(os.path.join(SCRIPT_DIR, 'README.rst'), encoding='utf-8') as readme:
        return readme.read()


def get_install_requirements():
    with io.open(os.path.join(SCRIPT_DIR, 'requirements-package.in'), encoding='utf-8') as reqs:
        return reqs.readlines()


def get_version():
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    with io.open(os.path.join(SCRIPT_DIR, 'oasislmf', '__init__.py'), encoding='utf-8') as init_py:
        return re.search('__version__ = [\'"]([^\'"]+)[\'"]', init_py.read()).group(1)


@contextmanager
def temp_dir():
    d = mkdtemp()
    yield d
    shutil.rmtree(d, ignore_errors=True)


version = get_version()
reqs = get_install_requirements()
readme = get_readme()


class PostInstallKtools(install):
    command_name = 'install'
    user_options = install.user_options + [
        ('ktools', None, 'Only install ktools components'),
    ]
    boolean_options = install.boolean_options + ['ktools']

    def __init__(self, *args, **kwargs):
        self.ktools_components = []
        install.__init__(self, *args, **kwargs)

    def run(self):
        self.install_ktools()
        install.run(self)

    def get_outputs(self):
        return install.get_outputs(self) + self.ktools_components

    def fetch_ktools_tar(self, location):
        self.announce('Retrieving ktools {}'.format(KTOOLS_VERSION), INFO)

        def _report(block_count, block_size, total_size, end='\r'):
            bar_size = 40
            filled = bar_size * (block_count * block_size) // total_size

            print('[{}{}]'.format('#' * filled, ' ' * (bar_size - filled)), end=end)

        urlretrieve('https://github.com/OasisLMF/ktools/archive/OASIS_{}.tar.gz'.format(KTOOLS_VERSION), location, _report)
        _report(1, 1, 1, end=' Done\n')

    def unpack_tar(self, tar_location, extract_location):
        self.announce('Unpacking ktools', INFO)
        with tarfile.open(tar_location) as tar:
            if not os.path.exists(extract_location):
                os.makedirs(extract_location)
            tar.extractall(extract_location)

    def build_ktools(self, extract_location):
        self.announce('Building ktools', INFO)
        build_dir = os.path.join(extract_location, 'ktools-OASIS_{}'.format(KTOOLS_VERSION))

        os.system('cd {build_dir} && ./autogen.sh && ./configure && make && make check'.format(build_dir=build_dir))
        return build_dir

    def add_ktools_to_path(self, build_dir):
        print('Installing ktools')

        for p in glob.glob(os.path.join(build_dir, 'src', '*', '*')):
            split = p.split(os.path.sep)

            # if the file name is the same as the directory we have found a
            # component executable
            if split[-1] == split[-2]:
                component_path = os.path.join(self.install_scripts, split[-1])
                shutil.copy(p, component_path)
                yield component_path

    def install_ktools(self):
        with temp_dir() as d:
            local_tar_path = os.path.join(d, 'ktools.tar.gz')
            local_extract_path = os.path.join(d, 'extracted')

            self.fetch_ktools_tar(local_tar_path)
            self.unpack_tar(local_tar_path, local_extract_path)
            build_dir = self.build_ktools(local_extract_path)
            self.ktools_components = list(self.add_ktools_to_path(build_dir))


# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


if sys.argv[-1] == 'publish':
    if os.system('pip freeze | grep twine'):
        print('twine not installed.\nUse `pip install twine`.\nExiting.')
        sys.exit()
    os.system('python setup.py sdist')
    os.system('twine upload dist/*')
    print('You probably want to also tag the version now:')
    print('  git tag -a {v} -m \'version {v}\''.format(v=version))
    print('  git push --tags')
    shutil.rmtree('dist')
    shutil.rmtree('build')
    shutil.rmtree('oasislmf.egg-info')
    sys.exit()


setup(
    name='oasislmf',
    version=version,
    packages=find_packages(exclude=('tests', 'tests.*', 'tests.*.*')),
    include_package_data=True,
    package_data={
        '': [
            'requirements-package.in',
            'LICENSE',
        ],
        'oasislmf/_data/': ['*']
    },
    exclude_package_data={
        '': ['__pycache__', '*.py[co]'],
    },
    scripts=['bin/oasislmf'],
    license='BSD 3-Clause',
    description='Core loss modelling framework.',
    long_description=readme,
    url='https://github.com/OasisLMF/oasislmf',
    author='Dan Bate,Sandeep Murthy',
    author_email="Dan Bate <dan.bate@wildfish.com>,Sandeep Murthy <sandeep.murthy@oasislmf.org>",
    keywords='oasis lmf loss modeling framework',
    install_requires=reqs,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    cmdclass={'install': PostInstallKtools},
)
